xquery version "3.1";

module namespace qm="https://mollenhauer-edition.de/ns/conversions/quotation-marks";

(:~  
 : Convert Module Quotation Marks
 : *******************************************************************************************
 : This is a module replacing Quotation-Marks-Elements with quotation marks.
 :  
 : @version 1.0 (2021-09-23)
 : @status stable
 : @author Max Zeterberg
 :)

declare namespace tei = "http://www.tei-c.org/ns/1.0";
declare namespace output="http://www.w3.org/2010/xslt-xquery-serialization";

declare option output:method "xml";

(:~
 : Default return for qm:quotation-marks: If $node is an elment node it will be
 : reconstructed and its child nodes will be passed on to qm:quotation-marks. 
 : If it is not an element, its child nodes will be passed on to qm:quotation-marks.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :  
 : @param $node The current node
:)
declare function qm:default-return($node as node()*)
{
    try {
        element {QName('http://www.tei-c.org/ns/1.0', $node/name())} {
            $node/@*,
            qm:quotation-marks($node/node())
        }
    } catch * {
        qm:quotation-marks($node/node())
    }
};

(:~
 : This function puts quotation marks before and after a node.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :
 : @param $node The current node
:)
declare function qm:create-qm($node as node()*, $qm as xs:string*)
{
    if (fn:matches($node/following-sibling::text()[1], '^\W+\s')) then
        (fn:concat(' ', $qm), qm:default-return($node), $qm)
    else
        (fn:concat(' ', $qm), qm:default-return($node), fn:concat($qm, ' '))
};

(:~
 : This function surrounds all quotation-elements (tei:q, tei:quote, tei:said, tei:u) with quotation marks.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0 
 :
 : @param $nodes TEI-Document without critical apparatus.
:)
declare function qm:quotation-marks($nodes as node()*) as item()*
{
    
    for $node in $nodes
    return
        typeswitch ($node)
            case text()
                return
                    ($node)
            case element(tei:q)
                return
                    if (($node/ancestor::tei:q) or ($node/ancestor::tei:quote) or ($node/ancestor::tei:said)) then
                        qm:create-qm($node, "'")
                    else
                        qm:create-qm($node, '"')
            case element(tei:quote)
                return
                    if (($node/ancestor::tei:q) or ($node/ancestor::tei:quote) or ($node/ancestor::tei:said)) then
                        qm:create-qm($node, "'")
                    else
                        qm:create-qm($node, '"')
            case element(tei:said)
                return
                    if (($node/ancestor::tei:q) or ($node/ancestor::tei:quote) or ($node/ancestor::tei:said)) then
                        qm:create-qm($node, "'")
                    else
                        qm:create-qm($node, '"')
            case element(tei:u)
                return
                    if (($node/ancestor::tei:q) or ($node/ancestor::tei:quote) or ($node/ancestor::tei:said)) then
                        qm:create-qm($node, "'")
                    else
                        qm:create-qm($node, '"')
            default
                return
                    qm:default-return($node)
};

(:~
 : Default return for qm:normalize-quotation-elements: If $node is an elment node it will be reconstructed 
 : and its child nodes will be passed on to qm:normalize-quotation-elements. If it is not an element, its 
 : child nodes will be passed on to qm:normalize-quotation-elements.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :  
 : @param $node The current node
:)
declare function qm:default-normalize-elements($node as node()*)
{
    try {
        element {QName('http://www.tei-c.org/ns/1.0', $node/name())} {
            $node/@*,
            qm:normalize-quotation-elements($node/node())
        }
    } catch * {
        qm:normalize-quotation-elements($node/node())
    }
};

(:~
 : This function normalizes the content of elements which have no child-elements.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :  
 : @param $node The current node
:)
declare function qm:normalize-elements($node as node()*)
{
    if ($node/element()) then
        qm:default-normalize-elements($node)        
    else
        element {QName('http://www.tei-c.org/ns/1.0', $node/name())} {
            $node/@*,
            normalize-space($node)
        }
};

(:~
 : This function normalizes the content of quotation-elements which have no child-elements.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :  
 : @param $nodes TEI-Document without critical apparatus.
:)
declare function qm:normalize-quotation-elements($nodes as node()*) as item()*
{
    for $node in $nodes
    return
        typeswitch ($node)
            case text()
                return
                    ($node)
            case element(tei:q)
                return
                    (qm:normalize-elements($node))
            case element(tei:quote)
                return
                    (qm:normalize-elements($node))
            case element(tei:said)
                return
                    (qm:normalize-elements($node))
            case element(tei:u)
                return
                    (qm:normalize-elements($node))
            default
                return
                    qm:default-normalize-elements($node)
};

(:~
 : Default return for qm:delete-quotation-elements: If $node is an elment node it will be
 : reconstructed and its child nodes will be passed on to qm:delete-quotation-elements. If 
 : it is not an element, its child nodes will be passed on to qm:delete-quotation-elements.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :  
 : @param $node The current node
:)
declare function qm:default-delete-quotation-elements($node as node()*)
{
    try {
        element {QName('http://www.tei-c.org/ns/1.0', $node/name())} {
            $node/@*,
            qm:delete-quotation-elements($node/node())
        }
    } catch * {
        qm:delete-quotation-elements($node/node())
    }
};

(:~
 : This function deletes the tags of quotation-elements. Other Elements will be reconstructed, Text will be normalized.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :  
 : @param $node The current node
:)
declare function qm:delete-element-tags($node as node()*)
{
    if ($node/element()) then
        for $childNode in $node/node()
        return
           if (local-name($childNode) = ('q', 'quote', "said", "u")) then
                qm:delete-quotation-elements($childNode)
            else
                try {
                    element {QName('http://www.tei-c.org/ns/1.0', $childNode/name())} {
                        $childNode/@*,
                        qm:delete-quotation-elements($childNode/node())
                    }
                } catch * {
                    normalize-space($childNode)
                }
    else
        $node/text()
};

(:~
 : This function deletes all quotation-elements.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :  
 : @param $nodes TEI-Document without critical apparatus.
:)
declare function qm:delete-quotation-elements($nodes as node()*) as item()*
{
    for $node in $nodes
    return
        typeswitch ($node)
            case text()
                return
                    ($node)
            case element(tei:q)
                return
                    (qm:delete-element-tags($node))
            case element(tei:quote)
                return
                    (qm:delete-element-tags($node))
            case element(tei:said)
                return
                    (qm:delete-element-tags($node))
            case element(tei:u)
                return
                    (qm:delete-element-tags($node))
            default
                return
                    qm:default-delete-quotation-elements($node)
};

(:~
 : This function calls all functions of this module in the right order.
 :
 : @author Max-Ferdinand Zeterberg
 : @version 1.0
 :  
 : @param $doc A TEI-Document.
 : @param $fassung The Fassung that should be converted to plain text.
:)
declare function qm:all($doc as node()*)
{
    qm:quotation-marks($doc)
    => qm:normalize-quotation-elements()
    => qm:delete-quotation-elements()
};