xquery version "3.1";

(:~
 : This module provides the API endpoints for various functions.
 :
 : @author Simon Sendler
 : @version 1.0
 : @since 0.4.1
 :)
module namespace kmg-rest="https://mollenhauer-edition.de/ns/kmg-rest";

import module namespace config="https://sade.textgrid.de/ns/config" at "../config.xqm";
import module namespace selection = "https://mollenhauer-edition.de/ns/conversions/selection" at "./conversions/selection.xqm";
import module namespace json = "https://mollenhauer-edition.de/ns/conversions/json" at "./conversions/json.xqm";
import module namespace rest="http://exquery.org/ns/restxq";
import module namespace functx="http://www.functx.com";

declare namespace tei="http://www.tei-c.org/ns/1.0";
declare namespace html="http://www.w3.org/1999/xhtml";
declare namespace output ="http://www.w3.org/2010/xslt-xquery-serialization";
declare namespace tg="http://textgrid.info/namespaces/metadata/core/2010";
declare namespace lido="http://www.lido-schema.org";

(:~
 : Provides the full xml of an edition.
 : 
 : @param $uri A valig TextGrid-URI without prefix
 : @return The requested XML-document
 :)
declare
  %rest:GET
  %rest:path("/api/full-xml/{$uri}")
  %output:method("xml")
function kmg-rest:full-xml($uri as xs:string)
as node()*
{
    let $doc-path := $config:sade-data || $uri || ".xml"
    return
        doc($doc-path)
};

(:~
 : Returns the combined output of the data- and metadata-functions
 :
 : @param $uri A valid TextGrid-URI without prefix
 : @return The requested information in JSON
 : @see get-metadata
 : @see get-data
 :)
declare 
  %rest:GET
  %rest:path("/api/{$uri}")
  %output:method("json")
function kmg-rest:get-all($uri as xs:string)
as map()*
{
    let $meta := kmg-rest:get-metadata($uri)
    let $data := kmg-rest:get-data($uri)
    return
        map {
            "metadata": $meta,
            "data": $data
        }
};

(: ~
 : Returns a map of the last-modification date of a TextGrid metadata object.
 :
 : @param $uri A TextGrid-URI
 : @return A map containing the last-modification date of the object as xs:dateTime and as human-readable string 
 :)
declare function kmg-rest:retrieve-transformation-date($uri as xs:string)
as map()
{
    let $date-last-transformed := xmldb:last-modified($config:sade-meta, $uri || ".xml")
    return
        try 
            {
                map 
                {
                    "string": kmg-rest:date-to-string($date-last-transformed),
                    "raw": $date-last-transformed
                }
            }
        catch * 
            {
                map { "error": "could not retrieve date of last transformation for object with URI " || $uri }
            }
};

(:~
 : Returns TextGrid-metadata of the requested document
 :
 : @param $uri A valid TextGrid-URI without prefix
 : @return The requested metadata in JSON 
 :)
declare
  %rest:GET
  %rest:path("/api/meta/{$uri}")
  %output:method("json")
function kmg-rest:get-metadata($uri as xs:string)
as map()*
{
    let $doc-path := $config:sade-meta || $uri || ".xml"
    let $doc := doc($doc-path)
    return
        kmg-rest:format-metadata($doc, $uri)
};

(: Formats TextGrid metadata for display via the API.
 :
 : @param $doc A TextGrid metadata-XML
 : @uri A TextGrid URI
 : @return A JSON object
 :)
declare function kmg-rest:format-metadata($doc as node()*, $uri as xs:string)
as map()*
{
    let $title := $doc//tg:title/text() => substring-after("-")
    let $date-created := 
        map {"string": kmg-rest:date-to-string($doc//tg:created/text()),
            "raw": $doc//tg:created/text()
        }
    let $date-last-modified := 
        map {"string": kmg-rest:date-to-string($doc//tg:lastModified/text()),
            "raw": $doc//tg:lastModified/text()
        }
    let $tg-uri := $doc//tg:textgridUri/text()
    let $output := 
        map {
            "title": $title,
            "created": $date-created,
            "modified": $date-last-modified,
            "transformed": kmg-rest:retrieve-transformation-date($uri),
            "tguri": $tg-uri
        }
    return
        $output
};

(:~
 : Returns formatted data in the teiHeader of the requested document.
 :
 : @param $uri A valid TextGrid-URI without prefix
 : @return The requested data in JSON 
 :)
declare
  %rest:GET
  %rest:path("/api/data/{$uri}")
  %output:method("json")
function kmg-rest:get-data($uri as xs:string)
as map()*
{
    let $doc-path := $config:sade-data || $uri || ".xml"
    let $doc := doc($doc-path)//tei:teiHeader
    return
        kmg-rest:format-data($doc)
};

(: ~
 : Formats relevant data about title, license and witnesses
 : in the teiHeader for display via the API.
 :
 : @param $doc a TEI-XML-document
 : @return A JSON object
 :)
declare function kmg-rest:format-data($doc as node()*)
as map()*
{
    let $licence := map {"name": normalize-space($doc//tei:licence/text()), "url": string($doc//tei:licence/@target)}
    let $titles :=
        for $title in $doc//tei:title return
            map {string($title/@type): normalize-space($title/text())}
    let $witnesses := 
        for $witness in $doc//tei:sourceDesc/tei:listWit/tei:witness return
            map {
                string($witness/@xml:id): map 
                {
                    "description": $witness/tei:desc/text(),
                    "type": if (contains($witness/tei:desc/@type, "Leittext")) then "main" else "version",
                    "year": functx:get-matches($witness/tei:desc, "\d{4}")[1]
                }
            }
    let $output := 
        map {
            "title": map:merge($titles),
            "licence": $licence,
            "witnesses": map:merge($witnesses)
        }
    return
        $output
};

(:~
 : This function turn a xs:dateTime into a string of the following format:
 : D?D.M?M.YYYY, H?H:M?M.
 : 
 : @param $date An xs:dateTime
 : @return A date string
 :)
declare function kmg-rest:date-to-string($date as xs:dateTime)
as xs:string
{
    let $day := string(fn:day-from-dateTime($date))
    let $month := string(fn:month-from-dateTime($date))
    let $year := string(fn:year-from-dateTime($date))
    let $hours := string(fn:hours-from-dateTime($date))
    let $minutes := string(fn:minutes-from-dateTime($date))
    let $date := $day || "." || $month || "." || $year
    let $time := $hours || ":" || $minutes
    let $date-string := $date || ", " || $time
    return
        $date-string
};

(:~
 : Returns the LIDO-information on the requested object
 :
 : @param $uri A valid TextGrid-URI without prefix
 : @return A JSON object
 :)
declare
  %rest:GET
  %rest:path("/api/lido/all")
  %output:method("json")
function kmg-rest:get-all-lido()
as map()*
{
    map:merge(for $doc in collection($config:lido) return
        let $uri := util:document-name($doc) => substring-before(".xml")
        return
            map {$uri: kmg-rest:get-single-lido($uri)})
};

(:~
 : Returns the LIDO-information on a requested object
 :
 : @param $uri A valid TextGrid-URI without prefix
 : @return The requested information in JSON
 :)
declare
  %rest:GET
  %rest:path("/api/lido/{$uri}")
  %output:method("json")
function kmg-rest:get-single-lido($uri as xs:string)
as map()
{
    let $doc-path := $config:lido || $uri || ".xml"
    let $doc := doc($config:lido || $uri || ".xml")
    return
        kmg-rest:format-lido($doc)
};

(: ~
 : Formats LIDO information on the requested object as a map
 :
 : @param $doc A LIDO-XML
 : @return A JSON object 
 :)
declare function kmg-rest:format-lido($doc)
as map()
{
    let $objectWorkType := $doc//lido:objectWorkType
    let $worktype := map 
        {
            "term": $objectWorkType/lido:term/text(),
            "href": $objectWorkType/lido:conceptID/text()
        }
    let $title :=  string-join($doc//lido:titleSet/lido:appellationValue/text(), ", ")
    let $repository := 
        if ($doc//lido:repositorySet) then
            for $repo in $doc//lido:repositorySet return
                map 
                    {
                        "name": $repo/lido:repositoryName//lido:appellationValue/text(),
                        "href": $repo//lido:legalBodyWeblink/text(),
                        "id": $repo/lido:workID/text(),
                        "location": for $location in $repo//lido:namePlaceSet return
                            $location//lido:appellationValue/text()
                    }
        else
            "unknown"
    let $descriptions :=
        for $description in $doc//lido:objectDescriptionSet return
            $description/lido:descriptiveNoteValue/text()
    let $measurements :=
        for $measurement in $doc//lido:objectMeasurementsSet return
            $measurement/lido:displayObjectMeasurements/text()

    let $creation := $doc//lido:event[lido:eventType/lido:term = "Daseinsbeginn"]
    let $artist-info := $creation/lido:eventActor
    let $artist := map 
        {
            "name": $artist-info/lido:displayActorInRole/text(),
            "href": $artist-info//lido:actorID[@lido:type = "http://terminology.lido-schema.org/identifier_type/uri"]/text()
        }
    let $place-of-creation := $creation//lido:displayPlace/text()
    let $year-of-creation := $creation//lido:displayDate/text()
    let $technique := $creation//lido:displayMaterialsTech/text()
    let $reproductions := $doc//lido:relatedWork//lido:objectID[@lido:type = "http://terminology.lido-schema.org/identifier_type/uri"]/text()
    let $external := for $description in $doc//lido:relatedWork[not(lido:objectID/@lido:type != "http://terminology.lido-schema.org/identifier_type/uri")] return
        $description//lido:objectWebResource/text()
    let $output :=  map 
        {
            "worktype": $worktype,
            "title": $title,
            "repository": $repository,
            "description": $descriptions,
            "measurements": $measurements,
            "artist": $artist,
            "origin": $place-of-creation,
            "year": $year-of-creation,
            "technique": $technique,
            "reproductions": $reproductions,
            "external": $external
        }
    return
        $output
};

(: ~
 : This functions allows to request the plain text of certain TEI-files as a JSON object.
 :
 : @param $title Title(s) of the texts
 : @param $year Year(s) in which the texts has been published the first time
 : @param $fassungen Fassung(en) which are requested
 : @return A JSON object 
 :)
declare
    %rest:path("api/plaintext")
    %rest:GET
    %rest:form-param("title", "{$title}")
    %rest:form-param("year", "{$year}")
    %rest:form-param("fassungen", "{$fassungen}")
    %output:media-type("application/json")
    %output:method("json")
function kmg-rest:plain_text($title as xs:string*, $year as xs:string*, $fassungen as xs:string*) 
as item()*
{
    let $filenames := selection:getFileNames($config:data-root, $title, $year)
    let $plain := json:plain-text($config:sade-data, $filenames, $fassungen)
    return
            $plain
};